// stdafx.cpp : source file that includes just the standard includes
// PyMI.pch will be the pre-compiled header
// stdafx.obj will contain the pre-compiled type information

#include "stdafx.h"
// Uncomment to enable memory leak detection
// Get VLD from: https://vld.codeplex.com/
//#include <vld.h>
